import json
import os

from classes.exceptions import DeploymentException
from .config_item import ConfigItem


class Config(object):
    """
    The Config object reads the config json named: "django-deploy.json" in the current working directory
    and initializes the config_item objects

    """
    CONFIG_JSON = 'django-deploy.json'
    CURRENT_WORKING_DIR = os.getcwd()
    CONFIG_FILE_PATH = os.path.join(CURRENT_WORKING_DIR, CONFIG_JSON)

    def __init__(self, config_file_path):
        self.config_file_path = config_file_path
        self.config_items = self.init_config_items()

    def config_item(self, deployment_type):
        """
        Searches a config item in the given config file

        :param deployment_type: the deployment type to search for
        :return: the found config item
        :rtype: ConfigItem
        """
        if deployment_type in self.config_items:
            return self.config_items[deployment_type]
        else:
            raise DeploymentException(
                'The deployment type in your django-deploy.json: {} was not found!\n'
                'Please read the docs to edit your django-deploy.json'.format(deployment_type)
            )

    def init_config_items(self):
        config_file = self.read_config_file(self.config_file_path)
        config_items = {}
        for key, values in config_file.items():
            config_items[key] = ConfigItem(key, values)
        return config_items

    def read_config_file(self, config_file_path=None):
        """
        :param config_file_path: the path of the config json
        :return: config_dictionary
        :rtype: dict
        """
        if not config_file_path:
            config_file_path = self.CONFIG_FILE_PATH
        try:
            return json.load(open(config_file_path))
        except Exception as exception:
            raise DeploymentException(str(exception))



import os

from fabric.api import *
from fabric.tasks import execute
from fabric.contrib.files import exists
from fabric.operations import put

from .exceptions import DeploymentException


class DeploymentServerActions(object):
    """
    The DeploymentServerActions class is a helper class for running shell commands on the server

    """

    DEPLOYED = 'deployed'
    PYTHON_2 = 'python'
    PYTHON_3 = 'python3'

    def __init__(self, config_item):
        """
        :param config_item:
        :type config_item: ConfigItem
        """
        self.config_item = config_item
        self.create_directory(self.config_item.deploy_dir)

    @property
    def deployed_path(self):
        """
        Concatination of deploy dir and deployed

        :return:
        """
        return os.path.join(self.config_item.deploy_dir, self.DEPLOYED)

    @property
    def virtualenv_path(self):
        """
        Concatination of virtualenv

        :return:
        """
        return os.path.join(self.config_item.deploy_dir, 'virtualenv')

    @property
    def python_path(self):
        """
        Concatination of virtualenv and python

        :return:
        """
        return os.path.join(os.path.join(self.virtualenv_path, 'bin'), 'python')

    @property
    def pip_path(self):
        """
        Concatination of virtulanev and pip

        :return:
        """
        return os.path.join(os.path.join(self.virtualenv_path, 'bin'), 'pip')

    def manage_py_full_path(self, to_deploy_commit_hash):
        """
        Full path of the manage.py in the newly deployed directory

        :param to_deploy_commit_hash:
        :return:
        """
        return os.path.join(
            os.path.join(
                self.config_item.deploy_dir,
                to_deploy_commit_hash
            ),
            self.config_item.manage_py.path
        )

    def _get_current_revision(self):
        """
        Gets the current deployed revision on the server

        :return:
        """
        if exists(self.deployed_path):
            return sudo('readlink {}'.format(self.deployed_path))
        else:
            return None

    def _execute_service_command(self, service, service_command):
        """
        Execute a command on a service through the service_management_command

        :param service: celery, gunicorn, wsgi...
        :param service_command: stop, start, reload, bla
        :return:
        """
        service_management_command = self.config_item.service_management_command

        if service_management_command == 'service':
            execute_string = '{service_management_command} {service} {command}'
        elif service_management_command == 'systemctl':
            execute_string = '{service_management_command} {command} {service}'
        else:
            raise DeploymentException(
                'Do you have a custom service running on your system? Please contact the django-deploy team.'
            )

        return sudo(
            execute_string.format(
                service_management_command=service_management_command,
                service=service,
                command=service_command,
            ),
            warn_only=True,
        )

    def _command_services_wrapper(self, command):
        """
        Executes the services in the config file

        :param command:
        :return:
        """
        for service in self.config_item.get_services_list():
            service_config_object = getattr(self.config_item.services, service)
            if command in service_config_object.__dict__:
                self.execute_with_host(
                    self._execute_service_command,
                    service=service,
                    service_command=getattr(service_config_object, command),
                )

    def _execute_manage_py_command(self, to_deploy_commit_hash, manage_py_command):
        """
        Executes the manage.py with the virtualenv and a specific command

        :param to_deploy_commit_hash:
        :param manage_py_command: migrate, collectstatic
        :return:
        """
        with cd(os.path.join(self.config_item.deploy_dir, to_deploy_commit_hash)):
            return sudo(
                '{python_virtualenv} {manage_py_full_path} {manage_py_command}'.format(
                    python_virtualenv='virtualenv/bin/python',
                    manage_py_full_path=self.config_item.manage_py.path,
                    manage_py_command=manage_py_command,
                )
            )

    def _set_symbolic_link(self, source_path, destination_path):
        """
        Set a symbolic link from source path to destination path
        ** Existing links will be deleted **

        :param source_path:
        :param destination_path:
        :return:
        """
        # Remove link if there is any
        self._remove_file_or_dir(destination_path)
        return sudo(
            'ln -sf {source_path} {destination_path}'.format(
                source_path=source_path,
                destination_path=destination_path,
            ),
            warn_only=True
        )

    def set_symbolic_link(self, source_path, destination_path):
        """
        See _set_symbolic_link

        :param source_path:
        :param destination_path:
        :return:
        """
        return self.execute_with_host(
            self._set_symbolic_link,
            source_path=source_path,
            destination_path=destination_path,
        )

    def execute_all_manage_py_commands(self, to_deploy_commit_hash):
        """
        Executes all manage.py commands in the config in the order of the list

        :param to_deploy_commit_hash:
        :return:
        """
        for manage_py_command in self.config_item.manage_py.actions:
            self.execute_with_host(
                self._execute_manage_py_command,
                to_deploy_commit_hash=to_deploy_commit_hash,
                manage_py_command=manage_py_command,
            )

    def stop_services(self):
        """
        Stops all services

        :return:
        """
        self._command_services_wrapper('stop')

    def start_services(self):
        """
        Starts all services

        :return:
        """
        self._command_services_wrapper('start')

    def has_virtualenv(self):
        """
        If the host has a virtualenv

        :return:
        """
        return self.execute_with_host(self._has_virtualenv).get(self.config_item.host)

    def update_virtualenv(self, to_deploy_commit_hash):
        """
        Updates virtualenv with the given requirements.txt in the config

        :param to_deploy_commit_hash:
        :return:
        """
        return self.execute_with_host(self._update_virtualenv, to_deploy_commit_hash=to_deploy_commit_hash)

    def _update_virtualenv(self, to_deploy_commit_hash):
        return sudo('{pip} install -r {requirements_txt}'.format(
            pip=self.pip_path,
            requirements_txt=os.path.join(
                os.path.join(
                    self.config_item.deploy_dir,
                    to_deploy_commit_hash
                ),
                self.config_item.requirements_txt
            )
        ))

    def create_virtualenv(self):
        return self.execute_with_host(self._create_virtualenv)

    def _create_virtualenv(self):
        """
        Creates the virtualenv in der deployment dir

        :return:
        """
        if self.config_item.python_version == self.PYTHON_2:
            return sudo('virtualenv {}'.format(self.virtualenv_path))
        elif self.config_item.python_version == self.PYTHON_3:
            return sudo('python3 -m venv {}'.format(self.virtualenv_path))

    def _has_virtualenv(self):
        """
        Looks for the virtualenv

        :return:
        """
        return exists(self.virtualenv_path, use_sudo=True)

    def _create_directory(self, directory):
        """
        Create the directory tree without overwriting the current tree (-p)

        :return:
        """
        sudo('mkdir -p {}'.format(directory))

    def _push_file_to_server(self, source_path, server_path):
        """
        Pushes the source file to server_path

        :return:
        """
        return put(source_path, server_path, use_sudo=True)

    def _unpack_tar_to_source(self, tar_path, unpack_to):
        """
        Unpacks a tar.gz to a source on the host

        :return:
        """
        return sudo('tar xf {tar_path} -C {unpack_to}'.format(tar_path=tar_path, unpack_to=unpack_to))

    def _remove_file_or_dir(self, source):
        """
        Removes a file or a directory on the server

        :param source:
        :return:
        """
        return sudo('rm -rf {}'.format(source))

    def remove_file_or_dir(self, source):
        """
        See _remove_file_or_dir

        :param source:
        :return:
        """
        self.execute_with_host(self._remove_file_or_dir, source=source)

    def unpack_tar_to_source(self, tar_path, unpack_to):
        """
        See _unpack_tar_to_source

        :param tar_path:
        :param unpack_to:
        :return:
        """
        return self.execute_with_host(self._unpack_tar_to_source, tar_path=tar_path, unpack_to=unpack_to)

    def push_file_to_server(self, source_path, server_path):
        """
        See _push_file_to_server

        :param source_path:
        :param server_path:
        :return:
        """
        return self.execute_with_host(self._push_file_to_server, source_path=source_path, server_path=server_path)

    def create_directory(self, directory):
        """
        See _create_directory

        :param directory:
        :return:
        """
        self.execute_with_host(self._create_directory, directory=directory)

    def get_current_revision(self):
        """
        See _get_current_revision

        :return:
        """
        return_value = self.execute_with_host(self._get_current_revision).get(self.config_item.host)
        return return_value.split('/')[-1] if return_value else return_value

    def execute_with_host(self, command, *args, **kwargs):
        """
        Wrapper function to execute a function with the current host

        :param command:
        :param args:
        :param kwargs:
        :return:
        """
        try:
            return execute(command, host=self.config_item.host, *args, **kwargs)
        except Exception as exception:
            raise DeploymentException(message=unicode(exception))



from fabric.api import local


class LocalGitWrapper(object):
    """
    The LocalGitWrapper object is a wrapper for commond git functions to run them locally on the machine

    """

    HEAD = 'HEAD'

    @staticmethod
    def get_branch_hash(branch):
        """
        Get the branch hash of given branch name

        :return:
        """
        return local('git rev-parse --verify {}'.format(branch), capture=True)

    @staticmethod
    def create_git_archive_of_branch(branch_hash, source_path):
        """
        Creates a git archive (tar.gz) in the source path

        :param branch_hash:
        :param source_path:
        :return:
        """
        local(
            'git archive {branch_hash} --format tar.gz -o {source_path}'.format(
                branch_hash=branch_hash,
                source_path=source_path,
            )
        )

    @staticmethod
    def remove_git_archive_of_branch(source_path):
        """
        Removes a git archive (or a file)

        :param source_path:
        :return:
        """
        local(
            'rm -rf {}'.format(source_path)
        )

    @staticmethod
    def get_differ_git_commit_messages(previous_commit_hash, next_commit_hash):
        """
        Gets the commit messages between two commits

        :param previous_commit_hash:
        :param next_commit_hash:
        :return:
        """
        return local("git log {}..{}".format(previous_commit_hash, next_commit_hash), capture=True)



from classes.exceptions import DeploymentException


class ConfigItem(object):
    """

    A deployment item describe a deployment type in the django-deploy.json

    Example:
        {
    --->  "prod": {
            "host": "google.com",
            "deploy_dir": "/var/www/test",
            "branch": "master"
          }
        }

    """

    def __init__(self, deployment_type, values):
        """
        Init the object with all values in the deployment type
        So the values like in the example from above would be:
        - host
        - deploy_dir
        - branch
        if the value is another dictionary it will be treated as another configitem with values

        :param deployment_type: the deployment type of current config item
        :param values: the dictionary of values
        """
        self.deployment_type = deployment_type
        for key, value in values.items():
            if isinstance(value, dict):
                setattr(self, key, ConfigItem(key, value))
            else:
                setattr(self, key, value)

    def __getattribute__(self, item):
        """
        Overwrite the getattribute function to raise the DeploymentException if an item wasn't found

        :param item:
        :return:
        """
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            raise DeploymentException(
                'Item \"{item}\" was not found in your django-deploy.json\n'
                'Please define the attribute in your deployment type: {deployment_type}'.format(
                    item=item,
                    deployment_type=self.deployment_type,
                )
            )

    def get_services_list(self):
        services_list = []
        for service in self.services.__dict__.keys():
            if service not in ['None', 'deployment_type']:
                services_list.append(service)
        return services_list



import sys

from setuptools import setup

if sys.version_info >= (3, 4):
    install_requires = ["fabric3==1.12.post1"]
else:
    install_requires = ["fabric==1.12.0"]

install_requires += ['click==6.6']

setup(
    name='django-deployment-sm',
    version="0.2.5",
    packages=['classes'],
    scripts=['django-deploy'],
    install_requires=install_requires,
    zip_safe=False,
)
